#include <ESP8266mDNS.h>
#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266HTTPUpdateServer.h>
#include <NeoPixelBus.h>
#include <Ticker.h>
#include "headlight.h"

/* ========================== NeoPixel  Section ============================= */
#define STRIP2 // remove this line to disable strip2

RgbColor amber(255, 80, 0);
RgbColor white(255,160,128);
RgbColor red(160, 0, 0);
RgbColor orange(140, 48, 0);
RgbColor yellow(120, 80, 0);
RgbColor green(0, 120, 0);
RgbColor blue(0,0,192);
RgbColor purple(160,0,192);
RgbColor black(0);

RgbColor cornerColor = amber;
volatile int cornerStripShow;
const uint16_t cornerLength = 18;
NeoPixelBus<NeoGrbFeature, NeoEsp8266Dma800KbpsMethod> cornerStrip(cornerLength);

#ifdef STRIP2
/* use a find & replace on "strip2_" to rename this strip however you like */
RgbColor strip2_Color = black;
volatile int strip2_StripShow;
const uint16_t strip2_Length = 18;
NeoPixelBus<NeoGrbFeature, NeoEsp8266Uart800KbpsMethod> strip2_Strip(strip2_Length);
#endif

/* synchronize Corner Strip with turn signal */
Ticker cornerTicker;
void turnISR() {
  cornerTicker.detach();
  
  int turnEdge = digitalRead(turnPin);
  /* turn signal just turned off */
  if (turnEdge) {
    for (int i = 0; i < cornerLength; i++) {
      cornerStrip.SetPixelColor(i, black);
    }
  }
  /* turn signal just turned on */ 
  else {
    for (int i = 0; i < cornerLength; i++) {
      cornerStrip.SetPixelColor(i, amber);
    }
  }
  cornerStripShow = 1;
  cornerTicker.once(1.25, turnCornerOn);
}

/* anti-synchronize strip2_ Eye with low beam */
void lowISR() {
#ifdef STRIP2
  int lowEdge = digitalRead(lowPin);
  /* low beam just turned off */
  if (lowEdge) {
    for (int i = 0; i < strip2_Length; i++) {
      strip2_Strip.SetPixelColor(i, strip2_Color);
    }
  } 
  /* low beam just turned on */
  else {
    for (int i = 0; i < strip2_Length; i++) {
      strip2_Strip.SetPixelColor(i, black);
    }
  }
  strip2_StripShow = 1;
#endif
}

/* Corner Strip Run at Startup */
void turnCornerOn() {
  for (int i = 0; i < cornerLength; i++) {
    cornerStrip.SetPixelColor(i, cornerColor);
  }
  cornerStripShow = 1;
}

/* strip2_ Eye Run at Startup */
void turnstrip2_On() {
  /* just run lowISR method */
  lowISR();
}

/* ========================== NeoPixel  Section ============================= */
/* ===========================  WiFi  Section  ============================== */
Ticker wifiTicker;

volatile int taskWiFicntr;
volatile int taskWiFichange;

void stopWiFi() {
  WiFi.disconnect();
  WiFi.mode(WIFI_OFF);
}
void startWiFi() {
  WiFi.mode(WIFI_STA);
  WiFi.setPhyMode(WIFI_PHY_MODE_11N);
  WiFi.setOutputPower(16);              // transmission power range 0 to +20.5 dBm
  WiFi.setAutoReconnect(HIGH);          // autoreconnect if connection lost
  WiFi.setSleepMode(WIFI_LIGHT_SLEEP);  // sets MCU & Radio to sleep mode when idle
  WiFi.hostname(WIFI_HOST_NAME);
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);
}
void checkWiFi() {
  if (taskWiFicntr > 20) {
    taskWiFichange = 1;
    taskWiFicntr = 0;
  }
  if (WiFi.status() == WL_CONNECTED) {
    wifiTicker.detach();
    wifiTicker.attach(15, checkWiFi);
    taskWiFicntr = 0;
  } else {
    taskWiFicntr++;
    wifiTicker.detach();
    wifiTicker.attach(1, checkWiFi);
    if (taskWiFicntr > 20) {
      taskWiFichange = 2;
      wifiTicker.detach();
      wifiTicker.attach(3, checkWiFi);
    }
  }
}
/* ===========================  WiFi  Section  ============================== */
/* =========================== Server  Section ============================== */
ESP8266WebServer server(80);
ESP8266HTTPUpdateServer httpUpdater;

void handleRoot() {
  server.send(200, "text/plain", WIFI_HELLO_STATEMENT);
}

void handleHelpMenu() {
  char *help_menu = "Help Menu:\n"
  "1) /help - displays this help page"
  "2) /setrgb?r=X&g=Y&b=Z - sets RGB color by Red = X, Green = Y, and Blue = Z\n"
  "3) /sethex?hex=0x000000 - sets RGB color by HEX value\n"
  "4) /gay - sets corner lights to rainbow color\n";
  server.send(200, "text/plain", help_menu);
  //Serial.print(help_menu);
}

void handleColorRGB() {
  uint8_t r = server.arg("r").toInt();
  uint8_t g = server.arg("g").toInt();
  uint8_t b = server.arg("b").toInt();

  cornerColor = RgbColor(r,g,b);
  
  for (int i = 0; i < cornerLength; i++) {
    cornerStrip.SetPixelColor(i, cornerColor);
  }
  cornerStripShow = 1;
}

void handleColorHEX() {
  char colorBuffer[8];
  server.arg(0).toCharArray(colorBuffer,8);
  uint32_t hexcolor = strtol(colorBuffer, NULL, 16);

  cornerColor = RgbColor(HtmlColor(hexcolor));
  
  for (int i = 0; i < cornerLength; i++) {
    cornerStrip.SetPixelColor(i, cornerColor);
  }
  cornerStripShow = 1;
}

void handleGayCorner() {
  cornerStrip.SetPixelColor(0, red);
  cornerStrip.SetPixelColor(1, red);
  cornerStrip.SetPixelColor(2, red);
  cornerStrip.SetPixelColor(3, amber);
  cornerStrip.SetPixelColor(4, amber);
  cornerStrip.SetPixelColor(5, amber);
  cornerStrip.SetPixelColor(6, yellow);
  cornerStrip.SetPixelColor(7, yellow);
  cornerStrip.SetPixelColor(8, yellow);
  cornerStrip.SetPixelColor(9, green);
  cornerStrip.SetPixelColor(10, green);
  cornerStrip.SetPixelColor(11, green);
  cornerStrip.SetPixelColor(12, blue);
  cornerStrip.SetPixelColor(13, blue);
  cornerStrip.SetPixelColor(14, blue);
  cornerStrip.SetPixelColor(15, purple);
  cornerStrip.SetPixelColor(16, purple);
  cornerStrip.SetPixelColor(17, purple);

  cornerStripShow = 1;
}

void handleNotFound() {
  String message = "File Not Found\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += (server.method() == HTTP_GET)?"GET":"POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";
  for (uint8_t i=0; i<server.args(); i++){
    message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
  }
  server.send(404, "text/plain", message);
  //Serial.print(message);
}

void startServer() {
  MDNS.begin(WIFI_HOST_NAME);
  
  server.on("/", handleRoot);
  server.on("/help", handleHelpMenu);
  server.on("/setrgb", handleColorRGB);
  server.on("/sethex", handleColorHEX);
  server.on("/gay", handleGayCorner);
  server.onNotFound(handleNotFound);
  server.begin();
}

void startOTA() {
  httpUpdater.setup(&server);
}

/* =========================== Server  Section ============================== */
/* ========================== Headlight Section ============================= */
void headlightInit() {
  taskWiFicntr = 0;
  taskWiFichange = 1;

  wifiTicker.attach(2, checkWiFi);
  
  strip2_StripShow = 0;
  strip2_Strip.Begin();
  cornerStripShow = 0;
  cornerStrip.Begin();
  
  turnCornerOn();
  turnstrip2_On();

  pinMode(lowPin, INPUT);
  pinMode(turnPin, INPUT);
  attachInterrupt(digitalPinToInterrupt(lowPin), lowISR, CHANGE);
  attachInterrupt(digitalPinToInterrupt(turnPin), turnISR, CHANGE);
}
/* ========================== Headlight Section ============================= */
/* ===========================  Main  Section  ============================== */
void setup() {
  //Serial.begin(74880);
  
  headlightInit();

  startServer();
  startOTA();
}

void loop() {
  if (taskWiFichange) {
    if (taskWiFichange == 1) { taskWiFichange = 0; startWiFi();}
    if (taskWiFichange == 2) { taskWiFichange = 0; stopWiFi(); }
  }
  if (cornerStripShow) {
    cornerStrip.Show();
    cornerStripShow = 0;
    server.send(200, "text/plain", "ok");
  }
  if (strip2_StripShow) {
    strip2_Strip.Show();
    strip2_StripShow = 0;
    server.send(200, "text/plain", "ok");
  }
  server.handleClient();
}
/* ===========================  Main  Section  ============================== */

